-- version and date
local revision = tonumber(("$Revision: 64559 $"):match("%d+"))
if not NBB_REVISION or revision > NBB_REVISION then
    NBB_REVISION = revision
    NBB_DATE = ("$Date: 2008-03-15 03:25:22 -0400 (Sat, 15 Mar 2008) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

-- localisation
local L = Rock("LibRockLocale-1.0"):GetTranslationNamespace("NBuffBars")
local LBS = LibStub("LibBabble-Spell-3.0")
local BS = LBS:GetLookupTable()
local BC = LibStub("LibBabble-Class-3.0"):GetLookupTable()

-- miscellaneous constants
NBB_HEADER_ID = "nbbheader"
NBB_DEFAULT_ICON = LBS:GetSpellIcon("Prayer of Fortitude")

-- get class information to determine which spells are castable
local _, CLASS = UnitClass("player")

-- presets for class/raid leader
NBB_PRESET_ORDER = { "RL", "PRIEST", "MAGE", "DRUID", "WARLOCK", "PALADIN", }
NBB_PRESET = {
    RL =      { name = L["Raid Leader"], byClass = true,     interesting = { "sta", "wil", "int", "ss", } ,},
    PRIEST =  { name = BC["Priest"],     byClass = false,    interesting = { "sta", "spi", "sp", "fw", }, },
    MAGE =    { name = BC["Mage"],       byClass = false,    interesting = { "int", "am", "dm", }, },
    DRUID =   { name = BC["Druid"],      byClass = false,    interesting = { "wil", "tho", }, },
    WARLOCK = { name = BC["Warlock"],    byClass = false,    interesting = { "ss", }, },
    PALADIN = { name = BC["Paladin"],    byClass = true,     interesting = { "bom", "bow", "bok", "bos", "bol", "boa", }, },
}

-- order that classes are shown, when grouping by class
NBB_CLASS_ORDER = {
    WARRIOR = 1,
    DRUID = 2,
    PALADIN = 3,
    PRIEST = 4,
    SHAMAN = 5,
    MAGE = 6,
    WARLOCK = 7,
    ROGUE = 8,
    HUNTER = 9,
    PET = 10,
}

-- mapping of CTRA buff indices to actual buffs
NBB_CTRA_0 = {}
NBB_CTRA_0[7] = "ss"
NBB_CTRA_0[9] = "tho"
NBB_CTRA_0[10] = "fw"
NBB_CTRA_0[20] = "am"
NBB_CTRA_0[21] = "dm"
NBB_CTRA_1_2 = {}
NBB_CTRA_1_2[1] = "sta"
NBB_CTRA_1_2[2] = "wil"
NBB_CTRA_1_2[3] = "int"
NBB_CTRA_1_2[5] = "sp"
NBB_CTRA_1_2[8] = "spi"
NBB_CTRA_1_2[11] = "bom"
NBB_CTRA_1_2[12] = "bow"
NBB_CTRA_1_2[13] = "bok"
NBB_CTRA_1_2[14] = "bos"
NBB_CTRA_1_2[15] = "bol"
NBB_CTRA_1_2[16] = "boa"

-- example data
NBB_EXAMPLE = {
    { name="Laughingcow", class="DRUID", group=1, remaining = { sta=3503, wil=3563, int=1701, spi=1736, bok=838, bol=288, tho=544, }, },
    { name="Maybor", class="WARRIOR", group=1, remaining = { sta=3503, wil=3563, bok=849, bol=874, tho=514, }, },
    { name="Pov", class="DRUID", group=1, remaining = { sta=3503, wil=3563, int=1725, spi=1705, bow=874, bok=838, }, },
    { name="Stimpy", class="PALADIN", group=1, remaining = { sta=3503, wil=3563, int=1752, spi=1676, bow=896, bok=864, }, },
    { name="Zaranth", class="WARRIOR", group=1, remaining = { sta=3503, wil=3563, bok=849, bol=874, tho=572, }, },
    { name="Baxter", class="PET", group=2, remaining = { sta=3572, wil=3588, int=3540, spi=3563, bom=841, bok=826, }, },
    { name="Damse", class="PRIEST", group=2, remaining = { sta=3572, wil=3588, int=3540, spi=3563, bow=852, bok=881, ss=1741, }, },
    { name="Esp", class="PRIEST", group=2, remaining = { sta=3572, wil=3588, int=3540, spi=3563, bok=881, bos=280, }, },
    { name="Nemes", class="PRIEST", group=2, remaining = { sta=3572, wil=3588, int=3540, spi=3563, bow=852, bok=881, ss=1669, }, },
    { name="Nyogtha", class="WARLOCK", group=2, remaining = { sta=3572, wil=3588, int=3540, spi=3563, bok=866, bos=859, }, },
    { name="Thootom", class="PET", group=2, remaining = { sta=1616, wil=1414, bom=841, bok=826, }, },
    { name="Trafalgar", class="HUNTER", group=2, remaining = { sta=3572, wil=3588, int=3540, spi=3563, bom=852, bok=859, }, },
    { name="Benten", class="ROGUE", group=3, remaining = { sta=3468, wil=3526, bom=845, bok=867, }, },
    { name="Falyme", class="ROGUE", group=3, remaining = { sta=3468, wil=3526, bom=845, bok=867, }, },
    { name="Malenga", class="SHAMAN", group=3, remaining = { sta=3468, wil=3526, int=1734, spi=1741, bow=890, bok=820, }, },
    { name="Theanimal", class="WARRIOR", group=3, remaining = { sta=1796, wil=3526, bom=252, bok=849, }, },
    { name="Woloomoo", class="DRUID", group=3, remaining = { sta=3468, wil=3526, spi=1745, bom=283, bok=838, }, },
    { name="Avyrn", class="MAGE", group=4, remaining = { sta=3552, wil=3432, int=3469, spi=3492, bok=825, bos=883, }, },
    { name="Cavalere", class="DRUID", group=4, remaining = { sta=3552, wil=3432, int=3469, spi=3492, bow=874, bos=296, }, },
    { name="Lemort", class="MAGE", group=4, remaining = { sta=3552, wil=3432, int=3469, spi=3492, bok=825, bos=883, }, },
    { name="Waxoff", class="MAGE", group=4, remaining = { sta=3552, wil=3432, int=3469, spi=3492, bok=825, bos=883, }, },
    { name="Yakkatak", class="SHAMAN", group=4, remaining = { sta=3552, wil=3432, int=3469, spi=3492, bow=890, bok=820, }, },
    { name="Ahlak", class="WARLOCK", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bok=866, bos=859, }, },
    { name="Angry", class="PET", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bom=841, bok=826, }, },
    { name="Constianni", class="PALADIN", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bow=896, bok=864, }, },
    { name="Geltuk", class="PET", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bom=841, bok=826, }, },
    { name="Kelendyn", class="HUNTER", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bom=852, bok=859, }, },
    { name="Lazrin", class="PET", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bom=841, bok=826, }, },
    { name="Supplement", class="WARLOCK", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bok=866, bos=859, }, },
    { name="WarKitten", class="PET", group=5, remaining = { sta=1437, wil=1584, bom=841, bok=826, }, },
    { name="Waxxon", class="HUNTER", group=5, remaining = { sta=3409, wil=3523, int=3584, spi=3263, bom=852, bok=859, }, },
}

-- information about the buffs that this addon is interested in
NBB_BUFF_ORDER = { "sta", "wil", "int", "spi", "bom", "bow", "bok", "bos", "bol", "boa", "sp", "ss", "tho", "fw", "am", "dm", }
NBB_BUFF = {
    sta = {
        name = L["Stamina"],
        duration = 3600,
        castable = (CLASS == "PRIEST"),
        icon = LBS:GetSpellIcon("Prayer of Fortitude"),
        single = BS["Power Word: Fortitude"],
        group = BS["Prayer of Fortitude"],
    },
    wil = {
        name = L["Wild"],
        duration = 3600,
        castable = (CLASS == "DRUID"),
        icon = LBS:GetSpellIcon("Gift of the Wild"),
        single = BS["Mark of the Wild"],
        group = BS["Gift of the Wild"],
    },
    int = {
        name = L["Intellect"],
        duration = 3600,
        castable = (CLASS == "MAGE"),
        icon = LBS:GetSpellIcon("Arcane Brilliance"),
        single = BS["Arcane Intellect"],
        group = BS["Arcane Brilliance"],
    },
    spi = {
        name = L["Spirit"],
        duration = 3600,
        castable = (CLASS == "PRIEST"),
        icon = LBS:GetSpellIcon("Prayer of Spirit"),
        single = BS["Divine Spirit"],
        group = BS["Prayer of Spirit"],
    },
    bom = {
        name = L["Might"],
        duration = 1800,
        castable = (CLASS == "PALADIN"),
        icon = LBS:GetSpellIcon("Greater Blessing of Might"),
        single = BS["Blessing of Might"],
        group = BS["Greater Blessing of Might"],
    },
    bow = {
        name = L["Wisdom"],
        duration = 1800,
        castable = (CLASS == "PALADIN"),
        icon = LBS:GetSpellIcon("Greater Blessing of Wisdom"),
        single = BS["Blessing of Wisdom"],
        group = BS["Greater Blessing of Wisdom"],
    },
    bok = {
        name = L["Kings"],
        duration = 1800,
        castable = (CLASS == "PALADIN"),
        icon = LBS:GetSpellIcon("Greater Blessing of Kings"),
        single = BS["Blessing of Kings"],
        group = BS["Greater Blessing of Kings"],
    },
    bos = {
        name = L["Salvation"],
        duration = 1800,
        castable = (CLASS == "PALADIN"),
        icon = LBS:GetSpellIcon("Greater Blessing of Salvation"),
        single = BS["Blessing of Salvation"],
        group = BS["Greater Blessing of Salvation"],
    },
    bol = {
        name = L["Light"],
        duration = 1800,
        castable = (CLASS == "PALADIN"),
        icon = LBS:GetSpellIcon("Greater Blessing of Light"),
        single = BS["Blessing of Light"],
        group = BS["Greater Blessing of Light"],
    },
    boa = {
        name = L["Sanctuary"],
        duration = 1800,
        castable = (CLASS == "PALADIN"),
        icon = LBS:GetSpellIcon("Greater Blessing of Sanctuary"),
        single = BS["Blessing of Sanctuary"],
        group = BS["Greater Blessing of Sanctuary"],
    },
    sp =  {
        name = L["Shadow Protection"],
        duration = 1200,
        castable = (CLASS == "PRIEST"),
        icon = LBS:GetSpellIcon("Prayer of Shadow Protection"),
        single = BS["Shadow Protection"],
        group = BS["Prayer of Shadow Protection"],
    },
    ss =  {
        name = L["Soulstone"],
        duration = 1800,
        castable = (CLASS == "WARLOCK"),
        icon = LBS:GetSpellIcon("Soulstone Resurrection"),
        single = BS["Soulstone Resurrection"],
        group = nil,
        item = L["Master Soulstone"],
    },
    tho = {
        name = L["Thorns"],
        duration = 600,
        castable = (CLASS == "DRUID"),
        icon = LBS:GetSpellIcon("Thorns"),
        single = BS["Thorns"],
        group = nil,
    },
    fw =  {
        name = L["Fear Ward"],
        duration = 600,
        castable = (CLASS == "PRIEST"),
        icon = LBS:GetSpellIcon("Fear Ward"),
        single = BS["Fear Ward"],
        group = nil,
    },
    am =  {
        name = L["Amplify Magic"],
        duration = 600,
        castable = (CLASS == "MAGE"),
        icon = LBS:GetSpellIcon("Amplify Magic"),
        single = BS["Amplify Magic"],
        group = nil,
    },
    dm =  {
        name = L["Dampen Magic"],
        duration = 600,
        castable = (CLASS == "MAGE"),
        icon = LBS:GetSpellIcon("Dampen Magic"),
        single = BS["Dampen Magic"],
        group = nil,
    },
}
