-- enhancements
--  option to remove pets or maybe just imps

-- code cleanup todo:
--  have a scheduled event to do updates rather than on event
--  look at remaining incombatlockdowns and possibly remove
--  optimise RaidMember:ScanBuffs and maybe use NBuffBars:UNIT_AURA
--  get rid of globals

-- version and date
local revision = tonumber(("$Revision: 64559 $"):match("%d+"))
if not NBB_REVISION or revision > NBB_REVISION then
    NBB_REVISION = revision
    NBB_DATE = ("$Date: 2008-03-15 03:25:22 -0400 (Sat, 15 Mar 2008) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

-- localisation
local L = Rock("LibRockLocale-1.0"):GetTranslationNamespace("NBuffBars")

-- constants
local ADDON_PATH = "Interface\\AddOns\\NBuffBars"
local TEXTURE_PATH = ADDON_PATH.."\\textures"
local BORDER_INSET = 2
local CLOSE_BUTTON_SCALE = 1.45
local CLOSE_BUTTON_OFFSET_X = 0.32
local RANGE_CHECK_FREQUENCY = 0.8
local BUFF_CHECK_FREQUENCY = 10

-- other libraries
local SharedMedia = LibStub("LibSharedMedia-2.0")
local RosterLib = AceLibrary("Roster-2.1")
local CandyBar = AceLibrary("CandyBar-2.0")

-- debug functions
function nbbError(msg)
    DEFAULT_CHAT_FRAME:AddMessage(tostring(msg), 1, 0, 0)
end
function nbbDebug(msg)
    DEFAULT_CHAT_FRAME:AddMessage(msg)
end

-- register the keybindings
BINDING_HEADER_NBUFFBARS = "NBuffBars"
BINDING_NAME_NBUFFBARS_SHOWHIDE = L["Show/Hide bars"]
BINDING_NAME_NBUFFBARS_TOGGLESORT = L["Toggle sorting by raid group/class"]
BINDING_NAME_NBUFFBARS_NEXTINTERESTING = L["Next interesting buff"]

-- create the class
NBuffBars = Rock:NewAddon("NBuffBars", "LibRockDB-1.0", "LibRockConfig-1.0", "LibRockHook-1.0", "LibRockEvent-1.0", "LibRockTimer-1.0", "LibFuBarPlugin-3.0")
NBuffBars.version = "r" .. NBB_REVISION
NBuffBars.revision = NBB_REVISION
NBuffBars.date = NBB_DATE

NBuffBars:SetDatabase("NBuffBarsDB")
NBuffBars:SetDefaultProfile("class")
NBuffBars:SetDatabaseDefaults("profile", {
    lock = false,
    visible = true,
    showInParty = true,
    showSolo = true,
    appearance = {
        colour = {
            byClass = true,
            frameBG = { 0, 0, 0, 0.5, },
            text = { 1, 1, 1, 1, },
            header = { 0.5, 0.5, 0.5, 1, },
            bar = { 0.7, 0.5, 1, 1,  },
            barBG = { 0.3, 0.2, 0.4, 1, },
            fade = 50,
        },
        byClass = false,
        group = true,
        texture = "Charcoal",
        textsize = 11,
        height = 14,
        width = 180,
        spacing = {
            bar = 1,
            group = 5,
        },
    },
    position = {
        x = nil,
        y = nil,
    },
    cur = nil,
    interesting = nil,
    boring = nil,
    cleanPreset = nil,
})

-- members
NBuffBars.member_m = nil
NBuffBars.group_m = nil
NBuffBars.exampleMember_m = nil
NBuffBars.dragging_m = false
NBuffBars.anchor_m = nil
NBuffBars.closeButton_m = nil
NBuffBars.bg_m = nil
NBuffBars.unitChangeCache_m = nil
NBuffBars.initialScan_m = nil
local debug = false

function NBuffBars:OnInitialize()
    if debug then nbbDebug("NBuffBars:OnInitialize") end

    self:SetupOptions()
    self:SetConfigSlashCommand("/nbuffbars", "/buffbars", "/bb", "/nbb")

    -- initialise tables
    self.group_m = {}

    local profile = self.db.profile
    local options = self.options

    -- Load status bar textures
    SharedMedia:Register(SharedMedia.MediaType.STATUSBAR, "Charcoal", TEXTURE_PATH.."\\Charcoal")
    SharedMedia:Register(SharedMedia.MediaType.STATUSBAR, "Cilo", TEXTURE_PATH.."\\Cilo")
    SharedMedia:Register(SharedMedia.MediaType.STATUSBAR, "Fourths", TEXTURE_PATH.."\\Fourths")
    SharedMedia:Register(SharedMedia.MediaType.STATUSBAR, "Healbot", TEXTURE_PATH.."\\Healbot")
    SharedMedia:Register(SharedMedia.MediaType.STATUSBAR, "LiteStep", TEXTURE_PATH.."\\LiteStep")
    SharedMedia:Register(SharedMedia.MediaType.STATUSBAR, "Round", TEXTURE_PATH.."\\Round")
    SharedMedia:Register(SharedMedia.MediaType.STATUSBAR, "Water", TEXTURE_PATH.."\\Water")

    -- create the anchor frame which is draggable - hidden by default
    self.anchor_m = CreateFrame("Button", "NBuffBarsAnchor", UIParent)
    self.anchor_m:EnableMouse(true)
    self.anchor_m:RegisterForDrag("LeftButton")
    self.anchor_m:SetClampedToScreen(true)
    self.anchor_m:Hide()
    self:UpdateLock()

    -- create the background frame - hidden by default
    self.bg_m = self.anchor_m:CreateTexture("NBuffBarsBG", "BACKGROUND")
    self.bg_m:SetTexture("Interface\\ChatFrame\\ChatFrameBackground")
    self.bg_m:SetPoint("TOPRIGHT", self.anchor_m, "TOPRIGHT", BORDER_INSET, BORDER_INSET)
    self.bg_m:Hide()
    self:UpdateBGColour()

    -- create the close button - hidden by default
    self.closeButton_m = CreateFrame("Button", "NBuffBarsCloseButton", self.anchor_m, "UIPanelCloseButton")
    self.closeButton_m:SetPoint("RIGHT", self.anchor_m, "RIGHT", 0, 0)
    self.closeButton_m:SetFrameStrata("HIGH")
    self.closeButton_m:Hide()

    -- populate the preset list
    self:PopulatePreset()

end

function NBuffBars:OnEnable(first)
    if debug then nbbDebug("NBuffBars:OnEnable") end

    -- see if oRA2 or CTRA is present
    if IsAddOnLoaded("oRA2") ~= 1 and IsAddOnLoaded("CT_RaidAssist") ~= 1 then
        nbbError(L["Warning: could not find oRA2 or CTRA; all raid members must have one of those installed for NBuffBars to function correctly!"])
    end

    -- WoW event registrations
    self:AddEventListener("PLAYER_LEAVING_WORLD")
    self:AddEventListener("PLAYER_REGEN_ENABLED")

    -- register the range check event
    self:AddRepeatingTimer("NBBCheckRange", RANGE_CHECK_FREQUENCY, "CheckRange")

    -- register the hooks for frame movement etc
    self:AddScriptHook(self.anchor_m, "OnDragStart", "AnchorDragStart")
    self:AddScriptHook(self.anchor_m, "OnDragStop", "AnchorDragStop")
    self:AddScriptHook(self.anchor_m, "OnMouseUp", "AnchorOnMouseUp")
    self:AddScriptHook(self.anchor_m, "OnEnter", "AnchorOnEnter")
    self:AddScriptHook(self.anchor_m, "OnLeave", "AnchorOnLeave")
    self:AddScriptHook(self.closeButton_m, "OnClick", "CloseButtonOnClick")

    -- update our appearance
    self:UpdateOptionTable()
    self:LoadAnchorPosition()
    self:OnUpdateFuBarText()
    
    -- we haven't built the set of bars yet
    self.initialScan_m = false

    -- wait until rosterlib is ready before scanning for raid members
    if first then
        self:AddEventListener("AceEvent-2.0", "AceEvent_FullyInitialized", "ScanRaid")
    else
        self:ScanRaid()
    end
end

function NBuffBars:OnDisable()
    if debug then nbbDebug("NBuffBars:OnDisable") end

    self:RemoveAllEventListeners()

    -- get rid of example bars / groups and don't bother updating the appearance afterwards
    self:HideExample(false)

    -- save the position of the anchor, in case some other frame modifier has moved it
    self:SaveAnchorPosition()

    -- hide the anchor and background
    self.anchor_m:Hide()
    self.bg_m:Hide()

    -- delete the members and groups
    if self.member_m then
        for k, v in pairs(self.member_m) do
            -- delete the bar
            v:Delete()
        end
    end
    self.member_m = nil
    self:BuildGroups()
end

-- scan all raid members and add them
function NBuffBars:ScanRaid()
    if debug then nbbDebug("NBuffBars:ScanRaid") end
    
    -- clear any existing members, just in case...
    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:Delete()
        end
    end

    -- process each unit, including pets
    for unit in RosterLib:IterateRoster(true) do
        self:RosterLib_UnitChanged(_, _, unit.unitid, unit.name, unit.class, unit.subgroup, _, nil, nil, nil, nil, _, true)
    end

    -- perform the updates that RosterLib_UnitChanged skipped in batch mode
    self:BuildGroups()
    self:UpdateVisibility()
    self:UpdateGeo()
    
    -- we've done the inital scan; any further updates are handled by RosterLib events
    self.initialScan_m = true

    -- subscribe to any future changes to roster or buffs
    self:AddEventListener("CHAT_MSG_ADDON")
    self:AddEventListener("AceEvent-2.0", "RosterLib_UnitChanged")
    self:AddEventListener("AceEvent-2.0", "SpecialEvents_UnitBuffGained")
    self:AddEventListener("AceEvent-2.0", "SpecialEvents_UnitBuffRefreshed")
    self:AddEventListener("AceEvent-2.0", "SpecialEvents_UnitBuffLost")
    
    -- check the auras every 30 seconds or so in case we miss an event or a refresh or other users don't have CTRA/oRA
    self:AddRepeatingTimer("NBBCheckBuff", BUFF_CHECK_FREQUENCY, "CheckBuffs")
end

function NBuffBars:PLAYER_LEAVING_WORLD(ns, event)

    -- save the position of the anchor, in case some other frame modifier has moved it
    self:SaveAnchorPosition()
end

-- called when the profile changed
function NBuffBars:OnProfileEnable()
    if debug then nbbDebug("NBuffBars:OnProfileEnable") end

    self:HideExample(false)

    -- the profile is stored in a new table, so we need to update it
    local appearance = self.db.profile.appearance
    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:SetAppearance(appearance)
        end
    end

    self:UpdateOptionTable()

    self:LoadAnchorPosition()

    self:UpdateVisibility()
    self:UpdateLock()
    self:UpdateTextures()
    self:UpdateTextSizes()
    self:UpdateBarPrefixes()
    self:UpdateHeader()
    self:UpdateBarColours()
    self:UpdateBarBGColours()
    self:UpdateTextColours()
    self:UpdateBGColour()
    self:UpdateGeo()

    CandyBar:SetCandyBarBackgroundColor(NBB_HEADER_ID, unpack(self.db.profile.appearance.colour.header))
end

-- called on profile reset
function NBuffBars:OnResetDatabase()
    if debug then nbbDebug("NBuffBars:OnResetDatabase") end
    
    self:OnProfileEnable()
end

function NBuffBars:AnchorDragStart()
    if InCombatLockdown() then return end

    if not self.db.profile.lock then
        self.dragging_m = true
        self:AnchorOnLeave()
        self.anchor_m:StartMoving()
        -- stop layout-cache.txt pollution
        self.anchor_m:SetUserPlaced(false)
    end

    self.hooks[self.anchor_m].OnDragStart()
end

function NBuffBars:AnchorDragStop()
    if InCombatLockdown() then return end

    self.dragging_m = false
    self.anchor_m:StopMovingOrSizing()

    self:SaveAnchorPosition()

    self.hooks[self.anchor_m].OnDragStop()
end

function NBuffBars:AnchorOnMouseUp(this, button)
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:AnchorOnMouseUp") end

    if button == "LeftButton" then
        self:NextInterestingBuff()
    elseif button == "RightButton" then
        self:OpenConfigMenu()
    end

    self.hooks[self.anchor_m].OnMouseUp(this, button)
end

function NBuffBars:AnchorOnEnter()
    if InCombatLockdown() or self.dragging_m then return end

    GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
    GameTooltip:ClearLines()
    GameTooltip:AddLine(L["|cffeda55fLeft-Click|r to monitor the next interesting buff.\n|cffeda55fRight-Click|r for options.\n|cffeda55fDrag|r to move all bars."])
    GameTooltip:Show()

    self.hooks[self.anchor_m].OnEnter()
end

function NBuffBars:AnchorOnLeave()
    GameTooltip:Hide()

    self.hooks[self.anchor_m].OnLeave()
end

function NBuffBars:CloseButtonOnClick()
    if debug then nbbDebug("NBuffBars:CloseButtonOnClick") end

    self:ToggleVisibleKey()
end

function NBuffBars:ToggleVisible()
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:ToggleVisible") end

    local profile = self.db.profile

    profile.visible = not profile.visible

    self:UpdateVisibility()
    self:UpdateHeader()
    self:UpdateGeo()
    self:UpdateFuBarTooltip()
end

function NBuffBars:LoadAnchorPosition()
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:LoadAnchorPosition") end

    local position = self.db.profile.position

    if not position.x or not position.y then
        self:ResetAnchorPosition()
    else
        self.anchor_m:ClearAllPoints()
        self.anchor_m:SetPoint("CENTER", UIParent, "CENTER", position.x, position.y)
    end
end

function NBuffBars:SaveAnchorPosition()
    if debug then nbbDebug("NBuffBars:SaveAnchorPosition") end

    local position = self.db.profile.position
    local x, y = self.anchor_m:GetCenter()
    x = x - GetScreenWidth() / 2
    y = y - GetScreenHeight() / 2
    position.x = x
    position.y = y
    self.anchor_m:ClearAllPoints()
    self.anchor_m:SetPoint("CENTER", UIParent, "CENTER", x, y)
end

function NBuffBars:ResetAnchorPosition()
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:ResetAnchorPosition") end

    -- move the anchor and save the new position
    self.anchor_m:ClearAllPoints()
    self.anchor_m:SetPoint("CENTER", UIParent, "CENTER", GetScreenWidth() * 1 / 6,  GetScreenHeight() * 1 / 3)
    self:SaveAnchorPosition()
end

-- receive any CTRA/oRA messages
function NBuffBars:CHAT_MSG_ADDON(ns, event, prefix, msg, type, author)
    if prefix ~= "CTRA" and prefix ~= "oRA" then return end
    if type ~= "RAID" then return end
    if debug then nbbDebug("NBuffBars:CHAT_MSG_ADDON "..tostring(prefix).." "..tostring(type).." "..tostring(author).." "..tostring(msg)) end
    self:SplitMessage(author, strsplit("#", msg))
end

-- copied from oRA
-- Iterate all the strings split from #
function NBuffBars:SplitMessage(author, ...)

    local updateNeeded = false

    for i = 1, select("#", ...) do
        local msg = select(i, ...)

        if debug then nbbDebug("NBuffBars:SplitMessage "..tostring(msg)) end

        -- decode any buff messages
        if string.match(msg, "^RN ") then
        
            local _, remaining, minor, major = strsplit(" ", msg)
            local buffTable = nil

            major = tonumber(major)
            minor = tonumber(minor)
            remaining = tonumber(remaining)

            if major == 0 then
                buffTable = NBB_CTRA_0
            elseif major == 1 or major == 2 then
                buffTable = NBB_CTRA_1_2
            end

            if remaining and minor and buffTable then

                local buffName = buffTable[minor]
                if buffName then

                    -- 1 second remaining indicates that the buff is not there; only process buff refreshes, as that is what SEA can't detect for buffs not cast by the player
                    if remaining > 1 and self.member_m then
                        local member = self.member_m[author]
                        if member then
                            if debug then nbbDebug(string.format(" updating %s %s %s", tostring(author), tostring(buffName), tostring(remaining))) end
                            member:SetBuffTime(buffName, nil, remaining)
                            updateNeeded = true
                        end
                    end
                end
            end
        end
    end

    -- change the order of the bars
    if updateNeeded then
        self:SortGroups()
        self:UpdateGeo()
    end
end

-- called repeatedly to check the range of the raid members
function NBuffBars:CheckRange()
    if self.member_m and self.db.profile.visible and self:CanShow() then
        for k, v in pairs(self.member_m) do
            v:CheckRange()
        end
    end
end

-- called repeatedly to check the status of buffs on all members
function NBuffBars:CheckBuffs()
    if debug then nbbDebug("NBuffBars:CheckBuffs") end
    
    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:ScanBuffs()
        end
    end
    
    self:SortGroups()
    self:UpdateGeo()
end

-- called when a raid member is added/changed/removed
function NBuffBars:RosterLib_UnitChanged(ns, event, unitId, name, class, group, _, oldName, oldunitId, oldClass, oldGroup, _, batchMode)
    if debug then nbbDebug("NBuffBars:RosterLib_UnitChanged "..tostring(unitId)..","..tostring(name)..","..tostring(class)..","..tostring(group).."  "..tostring(oldunitId)..","..tostring(oldName)..","..tostring(oldClass)..","..tostring(oldGroup).."  batchMode = "..tostring(batchMode)) end
    
    if InCombatLockdown() then
        if debug then nbbDebug(" caching change as in combat") end
        if not self.unitChangeCache_m then
            self.unitChangeCache_m = {}
        end
        table.insert(self.unitChangeCache_m, { unitId, name, class, group, oldName, oldunitId, oldClass, oldGroup, })
    else

        local member = nil
        local petChange = false

        if self.member_m then
            -- attempt to find this member
            if name and self.member_m then
                member = self.member_m[name]
            end
        else
            -- create the member table if necessary
            self.member_m = {}
        end

        -- add this member because they're new
        if not oldName and name and not member then
            if debug then nbbDebug(" adding "..tostring(name)) end
            if unitId and string.match(unitId, "pet") then
                petChange = true
            end
            member = self:AddMember(name, unitId, class, group)

        -- remove this member because they've left
        elseif not name and oldName then
            if debug then nbbDebug(" removing "..tostring(oldName)) end
            local oldMember = self.member_m[oldName]
            if oldMember then
                oldMember:Delete()
                self.member_m[oldName] = nil
            end

        -- remove the old and add the new because they have changed
        elseif name ~= oldName then
            if debug then nbbDebug(" changing "..tostring(oldName).." to "..tostring(name)) end
            local oldMember = self.member_m[oldName]
            if oldMember then
                oldMember:Delete()
                self.member_m[oldName] = nil
            end
            member = self:AddMember(name, unitId, class, group)

        -- change the group
        elseif group and group ~= oldGroup and member then
            if debug then nbbDebug(" changing group for "..tostring(name).." from "..tostring(oldGroup).." to "..tostring(group)) end
            member:SetGroup(group)

        -- change their raid ID
        elseif unitId and unitId ~= oldunitId and member then
            if debug then nbbDebug(" changing raid ID for "..tostring(name).." from "..tostring(oldunitId).." to "..tostring(unitId)) end
            member:SetUnitId(unitId)
        end
        
        -- when a party/raid member changes their pet, RosterLib does NOT remove the old pet
        -- this should be fixed in RosterLib
        if petChange and self.member_m then
            if debug then nbbDebug(" new pet "..tostring(name).." / "..tostring(unitId).."; scanning for old pet that needs to be removed") end
            for k, v in pairs(self.member_m) do
                if v.unitId_m == unitId and v.unitName_m ~= name then
                    local oldMember = self.member_m[v.unitName_m]
                    if oldMember then
                        if debug then nbbDebug(" removing old pet "..tostring(v.unitName_m)) end
                        oldMember:Delete()
                        self.member_m[v.unitName_m] = nil
                    end
                    break
                end
            end
        end

        -- sort the members into groups, or remove groups if necessary
        if not batchMode then
            self:BuildGroups()

            self:UpdateVisibility()
            self:UpdateGeo()
        end
    end
end

-- add the raid member
function NBuffBars:AddMember(name, unitId, class, group)
    if debug then nbbDebug("NBuffBars:AddMember adding "..tostring(name)) end
    
    if group < 1 or group > 8 then
        group = 1
    end

    local profile = self.db.profile

    local member = RaidMember:new(profile.appearance, name, unitId, class, group, profile.cur)
    self.member_m[name] = member
    if profile.visible and self:CanShow() then
        member:Show()
    end
    return member
end

-- called when we leave combat
function NBuffBars:PLAYER_REGEN_ENABLED(ns, event)
    if debug then nbbDebug("NBuffBars:PLAYER_REGEN_ENABLED") end

    -- process each cached change
    if self.initialScan_m and self.unitChangeCache_m then
        if debug then nbbDebug("NBuffBars:PLAYER_REGEN_ENABLED processing "..#self.unitChangeCache_m.." unit changes") end
        
        -- process each change in batch mode
        for i, v in ipairs(self.unitChangeCache_m) do
            self:RosterLib_UnitChanged(_, _, v[1], v[2], v[3], v[4], _, v[5], v[6], v[7], v[8], _, true)
        end

        -- perform the updates that RosterLib_UnitChanged skipped in batch mode
        self:BuildGroups()
        self:UpdateVisibility()
        self:UpdateGeo()

    -- build all bars, if we enabled the addon during combat
    elseif not self.initialScan_m then
        if debug then nbbDebug("NBuffBars:PLAYER_REGEN_ENABLED performing initial scan") end
        
        self:ScanRaid()
        self:LoadAnchorPosition()
    
    end
    
    -- clear the cache table
    self.unitChangeCache_m = {}
    self.unitChangeCache_m = nil
end

-- return true if the unit is a unit that matches our current party status i.e. solo, party, raid
function NBuffBars:IsUsefulUnit(unitId)

    -- always return player and player's pet
    if unitId == "player" or unitId == "pet" then
        return true
    -- return raid members if we're in a raid, but not player or player's pet
    elseif GetNumRaidMembers() > 0 then
        if string.match(unitId, "^raid") and not ((UnitIsUnit(unitId, "player") == 1) or (UnitIsUnit(unitId, "pet") == 1)) then
            return true
        end
    -- return party/partypets if we are in a party
    elseif GetNumPartyMembers() > 0 then
        if string.match(unitId, "^party") then
            return true
        end
    end
    
    return false
end

-- return the buff and duration from a localised name, nil if we're not scanning that buff
function NBuffBars:GetBuffFromName(buffName)
    for k, v in pairs(NBB_BUFF) do
        if v.single == buffName then
            return k, false, v.duration / 2
        elseif v.group == buffName then
            return k, true, v.duration
        end
    end
    return nil, nil, nil
end

-- update the buff time remaining
function NBuffBars:SetBuffTime(unitId, buff, isGroup, timeLeft)
    local unitName = UnitName(unitId)
    
    if self.member_m then
        local member = self.member_m[unitName]
        if member then
            member:SetBuffTime(buff, isGroup, timeLeft)
            self:SortGroups()
            self:UpdateGeo()
        end
    end
end

-- buff gained, lost or refreshed
-- function NBuffBars:UNIT_AURA(ns, event, unitId)
    -- if not self:IsUsefulUnit(unitId) then return end
    -- 
    -- if debug then nbbDebug(string.format("NBuffBars:UNIT_AURA(%s, %s, %s)", tostring(ns), tostring(event), tostring(unitId))) end
    -- 
    -- if member then
        -- local unitName = UnitName(unitId)
        -- if unitName then
            -- local member = self.member_m[unitName]
            -- if member then
                -- member:ScanBuffs()
            -- end
        -- end
    -- end
-- end

-- buff is gained; duration is unknown if buff was not cast by player
function NBuffBars:SpecialEvents_UnitBuffGained(ns, event, unitId, buffName, _, _, _, _, _, timeLeft)
    local buff, isGroup, duration = self:GetBuffFromName(buffName)
    if not self:IsUsefulUnit(unitId) or not buff then return end
    
    if debug then nbbDebug(string.format("NBuffBars:UnitBuffGained(%s, %s, %s)", tostring(unitId), tostring(buffName), tostring(timeLeft))) end
    
    if timeLeft then
        duration = timeLeft
    end
    self:SetBuffTime(unitId, buff, isGroup, duration)
end

-- buff is refreshed
function NBuffBars:SpecialEvents_UnitBuffRefreshed(ns, event, unitId, buffName, _, _, _, _, _, timeLeft)
    local buff, isGroup, duration = self:GetBuffFromName(buffName)
    if not self:IsUsefulUnit(unitId) or not buff then return end
    
    if debug then nbbDebug(string.format("NBuffBars:UnitBuffRefreshed(%s, %s, %s)", tostring(unitId), tostring(buffName), tostring(timeLeft))) end
    
    if timeLeft then
        duration = timeLeft
    end
    self:SetBuffTime(unitId, buff, isGroup, duration)
end

-- buff is lost
function NBuffBars:SpecialEvents_UnitBuffLost(ns, event, unitId, buffName)
    local buff = self:GetBuffFromName(buffName)
    if not self:IsUsefulUnit(unitId) or not buff then return end
    
    if debug then nbbDebug(string.format("NBuffBars:UnitBuffLost(%s, %s)", tostring(unitId), tostring(buffName))) end
    
    self:SetBuffTime(unitId, buff, nil)
end

-- put each of the members into groups, according to the current preferences
function NBuffBars:BuildGroups()
    if debug then nbbDebug("NBuffBars:BuildGroups") end

    local byClass = self.db.profile.appearance.byClass

    if self.exampleMember_m then
        member = self.exampleMember_m
    else
        member = self.member_m
    end

    -- clear the existing groups
    for i = 1, 10 do
        self.group_m[i] = {}
    end
    
    if member then
        -- put each of the members in a group
        local key, group
        for k, v in pairs(member) do
            if byClass then
                key = NBB_CLASS_ORDER[v.unitClass_m]
            else
                key = v.group_m
            end
            if debug then nbbDebug(" unit = "..tostring(v.unitName_m).." key = "..tostring(key)) end
            group = self.group_m[key]
            if group then
                table.insert(group, v)
            end
        end
        
        -- sort the groups
        self:SortGroups()
    end
end

function NBuffBars:ShowExample()
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:ShowExample") end

    if self.exampleMember_m then return end

    local example = self.options.args.example
    local profile = self.db.profile

    self.exampleMember_m = {}

    -- create a member for each example supplied
    local member = nil
    for i = 1, #NBB_EXAMPLE do
        member = ExampleRaidMember:new(profile.appearance, i, profile.cur)
        self.exampleMember_m[member.unitName_m] = member
        member:Show()
    end

    -- sort the members into groups
    self:BuildGroups()

    -- update the menu option
    example.name = L["Hide Example"]
    example.desc = L["Hide the example bars."]
    
    profile.visible = true

    self:UpdateVisibility()
    self:UpdateHeader()
    self:UpdateGeo()
end

function NBuffBars:HideExample(updateAppearance)
    if InCombatLockdown() then return end
    if not self.exampleMember_m then return end
    if debug then nbbDebug("NBuffBars:HideExample") end

    if updateAppearance == nil then
        updateAppearance = true
    end

    local example = self.options.args.example

    -- kill the example bars and groups
    for k, v in pairs(self.exampleMember_m) do
        v:Delete()
    end
    self.exampleMember_m = nil

    -- update the menu option and FuBar text
    example.name = L["Show Example"]
    example.desc = L["Show the example bars."]

    -- sort the members into groups
    self:BuildGroups()

    if updateAppearance then
        self:UpdateVisibility()
        self:UpdateHeader()
        self:UpdateGeo()
    end
end

-- return true if there are raid members or example members to show
function NBuffBars:CanShow()
    profile = self.db.profile
    return (GetNumRaidMembers() > 0 or profile.showInParty and GetNumPartyMembers() > 0 or profile.showSolo) and self.member_m ~= nil or self.exampleMember_m ~= nil
end

function NBuffBars:UpdateVisibility()
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:UpdateVisibility") end

    local profile = self.db.profile
    local appearance = profile.appearance
    local colour = appearance.colour

    -- only show if the user has asked for it and there are members to show
    if profile.visible and self:CanShow() then

        if not self.anchor_m:IsShown() then
            if debug then nbbDebug(" showing elements") end

            -- make elements visible
            self.anchor_m:Show()
            self.bg_m:Show()
            self.closeButton_m:Show()
            if not CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
                if debug then nbbDebug("  creating nbbheader") end
                CandyBar:RegisterCandyBar(NBB_HEADER_ID, 10, NBuffBars.name, DEFAULT_ICON, 0, 0, 0)
                CandyBar:SetCandyBarTimeLeft(NBB_HEADER_ID, 0)
                CandyBar:SetCandyBarFade(NBB_HEADER_ID, -1)
                CandyBar:StartCandyBar(NBB_HEADER_ID)
                CandyBar:SetCandyBarTexture(NBB_HEADER_ID, SharedMedia:Fetch(SharedMedia.MediaType.STATUSBAR, appearance.texture))
                CandyBar:SetCandyBarFontSize(NBB_HEADER_ID, appearance.textsize)
                CandyBar:SetCandyBarTextColor(NBB_HEADER_ID, unpack(colour.text))
                CandyBar:SetCandyBarColor(NBB_HEADER_ID, 0, 0, 0, 0)
                CandyBar:SetCandyBarBackgroundColor(NBB_HEADER_ID, unpack(colour.header))
            end
            if self.member_m then
                for k, v in pairs(self.member_m) do
                    v:Show()
                end
            end
            self:UpdateHeader()
        end
    else

        if self.anchor_m:IsShown() then
            if debug then nbbDebug(" hiding elements") end

            -- make elements invisible
            self.anchor_m:Hide()
            self.bg_m:Hide()
            self.closeButton_m:Hide()
            if CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
                CandyBar:UnregisterCandyBar(NBB_HEADER_ID)
            end
            if self.member_m then
                for k, v in pairs(self.member_m) do
                    v:Hide()
                end
            end
        end
    end

end

function NBuffBars:UpdateLock()
    if debug then nbbDebug("NBuffBars:UpdateLock") end

    self.anchor_m:SetMovable(not self.db.profile.lock)
end

function NBuffBars:UpdateTextures()
    if debug then nbbDebug("NBuffBars:UpdateTextures") end

    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:UpdateTexture()
        end
    end
    if self.exampleMember_m then
        for k, v in pairs(self.exampleMember_m) do
            v:UpdateTexture()
        end
    end
    if CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
        CandyBar:SetCandyBarTexture(NBB_HEADER_ID, SharedMedia:Fetch(SharedMedia.MediaType.STATUSBAR, self.db.profile.appearance.texture))
    end

end

function NBuffBars:UpdateTextSizes()
    if debug then nbbDebug("NBuffBars:UpdateTextSizes") end

    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:UpdateFontSize()
        end
    end
    if self.exampleMember_m then
        for k, v in pairs(self.exampleMember_m) do
            v:UpdateFontSize()
        end
    end
    if CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
        CandyBar:SetCandyBarFontSize(NBB_HEADER_ID, self.db.profile.appearance.textsize)
    end

end

function NBuffBars:UpdateBarPrefixes()
    if debug then nbbDebug("NBuffBars:UpdateBarPrefixes") end

    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:UpdatePrefix()
        end
    end
    if self.exampleMember_m then
        for k, v in pairs(self.exampleMember_m) do
            v:UpdatePrefix()
        end
    end
end

function NBuffBars:UpdateGeo()
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:UpdateGeo") end

    local appearance = self.db.profile.appearance
    local member = nil

    if self.exampleMember_m then
        member = self.exampleMember_m
    else
        member = self.member_m
    end

    -- set the size of the bars
    if member then
        for k, v in pairs(member) do
            -- allow a bit of overhang at the top of the select area if there's only a small gap
            if appearance.spacing.bar < 5 then
                v:SetSelectExtendTop(appearance.spacing.bar)
            else
                v:SetSelectExtendTop(0)
            end
            v:SetHeight(appearance.height)
            v:SetWidth(appearance.width, appearance.height)
        end
    end

    -- set width, adding: icon width = bar height
    self.anchor_m:SetWidth(appearance.width + appearance.height)
    self.bg_m:SetWidth(appearance.width + appearance.height + 2 * BORDER_INSET)
    self.closeButton_m:SetWidth(appearance.height * CLOSE_BUTTON_SCALE)
    
    -- set height
    if CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
        CandyBar:SetCandyBarWidth(NBB_HEADER_ID, appearance.width)
        CandyBar:SetCandyBarHeight(NBB_HEADER_ID, appearance.height)
    end
    self.anchor_m:SetHeight(appearance.height)
    self.closeButton_m:SetHeight(appearance.height * CLOSE_BUTTON_SCALE)

    -- shift the close button a bit to the right
    self.closeButton_m:SetPoint("RIGHT", self.anchor_m, "RIGHT", appearance.height * CLOSE_BUTTON_OFFSET_X, 0)

    -- store the vertical offset from the top of the anchor
    local y = 0 - appearance.height

    -- process each group
    for k, v in pairs(self.group_m) do

        -- skip empty groups
        if #v ~= 0 then

            -- add the group spacing
            y = y - appearance.spacing.group

            -- process each member in the group
            for l, w in ipairs(v) do

                -- add the bar spacing
                y = y - appearance.spacing.bar

                -- anchor the bar to the anchor's top right
                w:SetPoint("TOPRIGHT", self.anchor_m, "TOPRIGHT", 0, y)

                -- add the height of the bar itself
                y = y - appearance.height

            end
        end
    end

    -- positiong the header
    if CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
        CandyBar:SetCandyBarPoint(NBB_HEADER_ID, "TOPRIGHT", self.anchor_m, "TOPRIGHT")
    end

    -- size the background
    self.bg_m:SetHeight(0 - y + 2 * BORDER_INSET)

    -- move the real bars offscreen, if the examples are shown
    if self.exampleMember_m and self.member_m then
        for k, v in pairs(self.member_m) do
            v:SetPoint("TOPRIGHT", "UIParent", "BOTTOMLEFT", 0, 0)
        end
    end
end

-- update the header
function NBuffBars:UpdateHeader()
    if debug then nbbDebug("NBuffBars:UpdateHeader") end

    if CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
        local currentBuff = NBB_BUFF[self.db.profile.cur]
        CandyBar:SetCandyBarIcon(NBB_HEADER_ID, currentBuff.icon)
        CandyBar:SetCandyBarText(NBB_HEADER_ID, currentBuff.name)
    end
end

-- update bar colours
function NBuffBars:UpdateBarColours()
    if debug then nbbDebug("NBuffBars:UpdateBarColours") end

    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:UpdateBarColour()
        end
    end
    if self.exampleMember_m then
        for k, v in pairs(self.exampleMember_m) do
            v:UpdateBarColour()
        end
    end
end

-- update bar background colours
function NBuffBars:UpdateBarBGColours()
    if debug then nbbDebug("NBuffBars:UpdateBarBGColours") end

    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:UpdateBarBGColour()
        end
    end
    if self.exampleMember_m then
        for k, v in pairs(self.exampleMember_m) do
            v:UpdateBarBGColour()
        end
    end
end

-- update text colours
function NBuffBars:UpdateTextColours()
    if debug then nbbDebug("NBuffBars:UpdateTextColours") end

    -- header
    if CandyBar:IsCandyBarRegistered(NBB_HEADER_ID) then
        CandyBar:SetCandyBarTextColor(NBB_HEADER_ID, unpack(self.db.profile.appearance.colour.text))
    end

    -- bars
    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:UpdateTextColour()
        end
    end
    if self.exampleMember_m then
        for k, v in pairs(self.exampleMember_m) do
            v:UpdateTextColour()
        end
    end
end

-- update the background colour
function NBuffBars:UpdateBGColour()
    if debug then nbbDebug("NBuffBars:UpdateBGColour") end

    self.bg_m:SetVertexColor(unpack(self.db.profile.appearance.colour.frameBG))
end

-- set the buff we are to monitor
function NBuffBars:ChangeBuff(buff)
    if InCombatLockdown() then return end
    if debug then nbbDebug("NBuffBars:SetBuff "..tostring(buff)) end

    if not buff then return end

    self.db.profile.cur = buff

    -- change each of the bars to show the new buff
    if self.member_m then
        for k, v in pairs(self.member_m) do
            v:ChangeBuff(buff)
        end
    end
    if self.exampleMember_m then
        for k, v in pairs(self.exampleMember_m) do
            v:ChangeBuff(buff)
        end
    end

    self:UpdateHeader()
    self:UpdateFuBarText()
    self:SortGroups()
    self:UpdateGeo()
end

-- sort each group
function NBuffBars:SortGroups()
    if debug then nbbDebug("NBuffBars:SortGroups") end
    for k, v in pairs(self.group_m) do
        table.sort(v)
    end
end

-- move an entry from one integer indexed table to another
function NBuffBars:MoveEntry(entry, source, target)

    -- remove from the source table
    local found = false
    for i, v in ipairs(source) do
        if v == entry then
            table.remove(source, i)
            found = true
            break
        end
    end
    -- add to the target table
    if found then
        table.insert(target, entry)
    end

    self.db.profile.cleanPreset = nil

end

-- display the next buff on the interesting list
function NBuffBars:NextInterestingBuff()
    if InCombatLockdown() then return end

    local newBuff = nil
    local profile = self.db.profile

    -- get the first or next buff in the interesting list
    local currentIndex = nil
    for i, v in pairs(profile.interesting) do
        if v == profile.cur then
            currentIndex = i
        end
    end
    if not currentIndex or currentIndex == #profile.interesting then
        newBuff = profile.interesting[1]
    else
        newBuff = profile.interesting[currentIndex + 1]
    end

    -- update to show the new buff, if there is one to show
    if newBuff then
        self:ChangeBuff(newBuff)
    end

    self:UpdateHeader()
    self:UpdateFuBarText()
end

-- populate the option table with the list of buffs passed
function NBuffBars:BuildBuffList(opt, list, desc, func)

    local buff = nil
    local i = 101

    -- reset the list
    opt.args = {}

    -- re-insert each of the buffs
    for j, v in ipairs(list) do
        buff = NBB_BUFF[v]
        opt.args[buff.name] = {
            type = "execute",
            name = buff.name,
            desc = string.format(desc, buff.name),
            func = func,
            passValue = v,
            order = i,
        }
        i = i + 1
    end
end

-- load the profile with the information for the preset specified
function NBuffBars:LoadPreset(class)
    if debug then nbbDebug("NBuffBars:LoadPreset class = "..tostring(class)) end

    local preset = NBB_PRESET[class]
    local profile = self.db.profile

    -- use raid leader if we can't find the class specified
    if not preset then
        if debug then nbbDebug("NBuffBars: using RL list as there is no preset for this class") end
        preset = NBB_PRESET.RL
        class = "RL"
    end

    -- clear the existing lists
    profile.interesting = {}
    profile.boring = {}

    -- populate the lists based on the preset
    local found
    for i, v in ipairs(NBB_BUFF_ORDER) do
        found = false

        for j, w in ipairs(preset.interesting) do
            if v == w then
                table.insert(profile.interesting, v)
                found = true
            end
        end

        if not found then
            table.insert(profile.boring, v)
        end
    end

    -- set the current to the first in the interesting list
    profile.cur = profile.interesting[1]

    -- set the group by
    profile.appearance.byClass = preset.byClass

    -- this profile is currently being used
    profile.cleanPreset = class

    -- set the new buff
    self:ChangeBuff(profile.cur)

    -- sort the members into groups
    self:BuildGroups()

    self:UpdateHeader()
    self:UpdateGeo()
end

-- toggle betweeen sorting by group and by class
function NBuffBars:ToggleSort()
    if InCombatLockdown() then return end

    local profile = self.db.profile

    profile.appearance.byClass = not profile.appearance.byClass

    -- update the menus
    self:UpdateOptionTable()

    -- rebuild the groups
    self:BuildGroups()

    profile.cleanPreset = nil

    self:UpdateGeo()
end

-- toggle visible keybinding pressed
function NBuffBars:ToggleVisibleKey()
    if InCombatLockdown() then return end

    if self:CanShow() then
        if self.exampleMember_m then
            self:HideExample()
        else
            self:ToggleVisible()
        end
    end
end

-- next interesting keybinding pressed
function NBuffBars:NextInterestingKey()
    if InCombatLockdown() then return end

    self:NextInterestingBuff()
end

-- update the option table to reflect the user's current preferences
local function MakeInteresting(buff)
    NBuffBars:MoveEntry(buff, NBuffBars.db.profile.boring, NBuffBars.db.profile.interesting)
    NBuffBars:UpdateOptionTable()
end
local function MakeBoring(buff)
    NBuffBars:MoveEntry(buff, NBuffBars.db.profile.interesting, NBuffBars.db.profile.boring)
    NBuffBars:UpdateOptionTable()
end
function NBuffBars:UpdateOptionTable()

    local optionargs = self.options.args
    local profile = self.db.profile

    -- initialise interesting/boring lists if they haven't been
    if not profile.interesting or not profile.boring then
        local _, class = UnitClass("player")
        self:LoadPreset(class)
    end

    -- set the group by item
    if profile.appearance.byClass then
        optionargs.groupBy.name = string.format(L["Group By %s"], L["Raid Group"])
        optionargs.groupBy.desc = string.format(L["Arrange the bars in groups by |cffffffff%s|r."], L["Raid Group"])
    else
        optionargs.groupBy.name = string.format(L["Group By %s"], L["Class"])
        optionargs.groupBy.desc = string.format(L["Arrange the bars in groups by |cffffffff%s|r."], L["Class"])
    end

    -- set the boring / interesting change options
    self:BuildBuffList(optionargs.makeInteresting, profile.boring, L["Move |cffffffff%s|r to the interesting buff list."], MakeInteresting)
    self:BuildBuffList(optionargs.makeBoring, profile.interesting, L["Move |cffffffff%s|r to the boring buff list."], MakeBoring)
end

-- populate the preset list
local function SelectPreset(class)
    if debug then nbbDebug("SelectPreset class = "..tostring(class)) end
    NBuffBars:LoadPreset(class)
    NBuffBars:UpdateOptionTable()
end
function NBuffBars:PopulatePreset()

    local presetargs = self.options.args.preset.args
    local groupBy = nil
    local buffs = nil
    local preset = nil
    local j = 1

    for k, _ in pairs(NBB_PRESET) do

        preset = NBB_PRESET[k]

        if preset.byClass then
            groupBy = L["Class"]
        else
            groupBy = L["Raid Group"]
        end

        -- build a list of buffs that this preset is interested in
        for l, w in pairs(preset.interesting) do
            if buffs then
                buffs = buffs..", "..NBB_BUFF[w].name
            else
                buffs = NBB_BUFF[w].name
            end
        end

        -- populate the options with each of the groupings
        presetargs[preset.name] =  {
            type = "toggle",
            name = preset.name,
            desc = string.format(L["Group bars by |cffffffff%s|r.\nMonitor the following buffs: |cffffffff%s|r."], groupBy, buffs),
            get = function() return NBuffBars.db.profile.cleanPreset == k end,
            set = SelectPreset,
            passValue = k,
            order = j,
        }

        buffs = nil

        j = j + 1
    end
end
